// Copyright, Simon Halphen Boserup, 2022

/* The _dstpctile_fast is built as an analogue to Stata's built-in procedue _pctile. 
   It comes in two versions:
   
   The procedure comes in two versions:
   - _dstpctile
   - _dstpctile_fast
   The latter DOES NOT PRESERVE SORT ORDER of the dataset. The benefit is that 
   the procedure runs an order of magnitude faster, which is helpful when using 
   large datasets.
   
   Syntax:
   _dstpctile_fast varname, percentiles(numlist) [minobs(integer) detail]
   
   numlist = percentages (e.g., 50 for the median).

   
   Functionality:
	- Can take list of percentiles like _pctile.
	- Locates observation closest to wanted percentile and takes a mean over 
	  this and a number of adjacent observations to either side.
	- Option "minobs" chooses the total number of observations used (default: 5).
	  Note: THERE IS NO EXPLICIT INTERPOLATION. Selecting minobs(1) will only provide
	  the same result as _pctile if the percentile happens to match exactly with
	  one particular observation.
	- Not byable.
	
	Output can be listed by running 'return list'.
	
	Standard output:
	 - r(r1) .. r(rN)		: percentile values
	 
	Detail output:
	 - rj_min				: min value used in the mean for percentile no. j
	 - rj_max				: max value used in the mean for percentile no. j
	 - rj_dif				: difference between the max and the min
	 - obsForMean			: number of observations used in average for the pseudo percentile
	 
*/

capture program drop _dstpctile_fast
program define _dstpctile_fast , rclass 
	version 16
	syntax varlist(max=1) [if] [in] , Percentiles(string)  [Minobs(integer 5) Detail]
	
	loc var : word 1 of `varlist'
	
	// sort order may not be preserved
	sort `1'
	
	
	// setup
	marksample touse

	token `percentiles'
	loc maxperc : word count `percentiles'
	
	
	// percentile list in matrix q	
	mata: q = J(1,`maxperc',.)
	forval i=1/`maxperc' {
	    mata: q[1,`i'] = ``i''/100
	}
	//mata: q
	
	
	// get index number of the minobs # of observations closest in rank to particular percentile
	qui count if `touse'
	loc N = r(N)
	
	mata: offset = J(1,`maxperc',0)
	//mata: offset
	local first_time = 1
	local dom = 1
	
	if `N' > 4 {
		while `dom' == 0 | `first_time' == 1{
			
			local first_time = 0
		//make sure it uses exactly 5 obs
		
			/*if `dom' == 0{
				local offset = `offset' +1
				//di "`offset'"
			}*/
			
			local dom = 1
			
			mata: q = mm_cond(q :> (`N'-2)/`N', (`N'-2)/`N', q)
			mata: q = mm_cond(q :< 3/`N', 3/`N', q)
			//mata: q*`N'
			
			loc bounds = (`minobs'-1)/2
			
			mata: lo = round(`N'*q:-`bounds') - offset // lower bound indices for each percentile
			mata: hi = round(`N'*q:+`bounds') - offset // upper bound
			
			//mata: lo
			//mata: hi
			
			mata: obsIndex = (lo[1,1]::hi[1,1]) // col vec of index numbers for first percentile
			if `maxperc'>1 { // append cols for each additional percentile requested
				forval i=2/`maxperc' {
					mata: obsIndex = (obsIndex, (lo[1,`i']::hi[1,`i']) )
				}
			}
			//mata: obsIndex
			
			// print no. of observations used for calculations
			mata: o=rows(obsIndex)
			mata: st_numscalar("o",o)
			//di as text "Percentiles are calculated as means over " o " observations."	
				
			
			// calculate means around percentiles
			mata: st_view(X=.,.,"`var'", "`touse'") // create view of Stata variable in question
				
			mata: values = J(rows(obsIndex), `maxperc',.) // matrix for var values at relevant indices
			forval i=1/`maxperc'{
				mata: values[.,`i'] = X[obsIndex[.,`i']]
				
			}	
			mata: max_ = colmax(values) 
			mata: sum_ = colsum(values) 
			mata: frac_ = max_ :/ sum_
			mata: st_matrix("S_frac",frac_)
			
			
			forvalues ii = 1/`maxperc'{
				if S_frac[1,`ii']>0.4{
					local dom = 0
					mata: offset[1,`ii'] = offset[1,`ii']+1
					di "Violator"
					mata: offset
				}
				
			}
			
			
		}
			
		//mata: values
		mata: output = mean(values) // calculate means
		//mata: output
		mata: st_matrix("results",output)
			
			
		// report results
		* backward loop ensures return list printed in correct order
		
		* detail info
		if "`detail'" != "" {
			mata: minval = values[1,.]
			mata: maxval = values[rows(values),.]
			mata: dif = maxval :- minval
			
			mata: st_matrix("minval",minval)
			mata: st_matrix("maxval",maxval)
			mata: st_matrix("dif",dif)
			
			forval i = `maxperc'(-1)1 {
				return scalar r`i'_dif = dif[1,`i']
				return scalar r`i'_max = maxval[1,`i']
				return scalar r`i'_min = minval[1,`i']
			}
			
			return scalar obsForMean = o
		}
		
		
		
		* standard output
		forval i = `maxperc'(-1)1 {
			return scalar r`i' = results[1,`i']
			
		}
		mata: st_numscalar("r(sumoffset)",sum(offset))
		mata: index = select((1..cols(offset)),colmax(offset:>0))
		mata: st_numscalar("r(index)",sum(index))
		return add
		
	}
	
end
